import tensorflow as tf
import logging
import json
import os

from sklearn.metrics import precision_recall_fscore_support, accuracy_score
from tensorflow.keras.layers import LSTM, Dense, BatchNormalization, LayerNormalization
from tensorflow.keras.layers import Input, Dense
from tensorflow.keras import Model
from core.astromer import get_ASTROMER
from core.metrics import custom_acc
from core.tboard import save_scalar
from core.losses import custom_bce
from core.output import SauceLayer
from tqdm import tqdm

from core.data import standardize
logging.getLogger('tensorflow').setLevel(logging.ERROR)  # suppress warnings

def get_fc_attention(units, num_classes, weigths):
    ''' FC + ATT'''

    conf_file = os.path.join(weigths, 'conf.json')
    with open(conf_file, 'r') as handle:
        conf = json.load(handle)

    model = get_ASTROMER(num_layers=conf['layers'],
                         d_model   =conf['head_dim'],
                         num_heads =conf['heads'],
                         dff       =conf['dff'],
                         base      =conf['base'],
                         dropout   =conf['dropout'],
                         maxlen    =conf['max_obs'],
                         use_leak  =conf['use_leak'])
    weights_path = '{}/weights'.format(weigths)
    model.load_weights(weights_path)
    encoder = model.get_layer('encoder')
    encoder.trainable = False

    mask = 1.-encoder.input['mask_in']
    x = encoder(encoder.input)
    x = x * mask
    x = tf.reduce_sum(x, 1)/tf.reduce_sum(mask, 1)

    x = Dense(1024, name='FCN1')(x)
    x = Dense(512, name='FCN2')(x)
    x = Dense(num_classes, name='FCN3')(x)
    return Model(inputs=encoder.input, outputs=x, name="FCATT")

def get_lstm_no_attention(units, num_classes, maxlen, dropout=0.5):
    ''' LSTM + LSTM + FC'''

    serie  = Input(shape=(maxlen, 1),
                  batch_size=None,
                  name='input')
    times  = Input(shape=(maxlen, 1),
                  batch_size=None,
                  name='times')

    mask   = Input(shape=(maxlen, 1),
                  batch_size=None,
                  name='mask')
    length = Input(shape=(maxlen,),
                  batch_size=None,
                  dtype=tf.int32,
                  name='length')

    placeholder = {'input':serie,
                   'mask_in':mask,
                   'times':times,
                   'length':length}

    bool_mask = tf.logical_not(tf.cast(placeholder['mask_in'], tf.bool))

    x = tf.concat([placeholder['times'], placeholder['input']], 2)

    x = LSTM(units, return_sequences=True,
             dropout=dropout, name='RNN_0')(x, mask=bool_mask)
    x = LayerNormalization(axis=1)(x)
    x = LSTM(units, return_sequences=True,
             dropout=dropout, name='RNN_1')(x, mask=bool_mask)
    x = LayerNormalization(axis=1)(x)
    x = Dense(num_classes, name='FCN')(x)

    return Model(inputs=placeholder, outputs=x, name="RNNCLF")

def get_lstm_attention(units, num_classes, weigths, dropout=0.5):
    ''' ATT + LSTM + LSTM + FC'''
    conf_file = os.path.join(weigths, 'conf.json')
    with open(conf_file, 'r') as handle:
        conf = json.load(handle)

    model = get_ASTROMER(num_layers=conf['layers'],
                         d_model   =conf['head_dim'],
                         num_heads =conf['heads'],
                         dff       =conf['dff'],
                         base      =conf['base'],
                         dropout   =conf['dropout'],
                         maxlen    =conf['max_obs'],
                         use_leak  =conf['use_leak'])
    weights_path = '{}/weights'.format(weigths)
    model.load_weights(weights_path)
    encoder = model.get_layer('encoder')
    encoder.trainable = False

    bool_mask = tf.logical_not(tf.cast(encoder.input['mask_in'], tf.bool))

    x = encoder(encoder.input)
    x = tf.reshape(x, [-1, conf['max_obs'], encoder.output.shape[-1]])
    x = LayerNormalization()(x)

    x = LSTM(units, return_sequences=True,
             dropout=dropout, name='RNN_0')(x, mask=bool_mask)
    x = LayerNormalization()(x)
    x = LSTM(units, return_sequences=True,
             dropout=dropout, name='RNN_1')(x, mask=bool_mask)
    x = LayerNormalization()(x)
    x = Dense(num_classes, name='FCN')(x)

    return Model(inputs=encoder.input, outputs=x, name="RNNCLF")

@tf.function
def train_step(model, batch, opt):
    with tf.GradientTape() as tape:
        y_pred = model(batch)
        ce = custom_bce(y_true=batch['label'], y_pred=y_pred)
        acc = custom_acc(batch['label'], y_pred)
    grads = tape.gradient(ce, model.trainable_weights)
    opt.apply_gradients(zip(grads, model.trainable_weights))
    return acc, ce

@tf.function
def valid_step(model, batch, return_pred=False):
    with tf.GradientTape() as tape:
        y_pred = model(batch, training=False)
        ce = custom_bce(y_true=batch['label'],
                         y_pred=y_pred)
        acc = custom_acc(batch['label'], y_pred)
    if return_pred:
        return acc, ce, y_pred, batch['label']
    return acc, ce

def train(model,
          train_batches,
          valid_batches,
          patience=20,
          exp_path='./experiments/test',
          epochs=1,
          lr=1e-3,
          verbose=1):
    # Tensorboard
    train_writter = tf.summary.create_file_writer(
                                    os.path.join(exp_path, 'logs', 'train'))
    valid_writter = tf.summary.create_file_writer(
                                    os.path.join(exp_path, 'logs', 'valid'))
    # Optimizer
    optimizer = tf.keras.optimizers.Adam(lr)
    # To save metrics
    train_bce  = tf.keras.metrics.Mean(name='train_bce')
    valid_bce  = tf.keras.metrics.Mean(name='valid_bce')
    train_acc  = tf.keras.metrics.Mean(name='train_acc')
    valid_acc  = tf.keras.metrics.Mean(name='valid_acc')

    # ==============================
    # ======= Training Loop ========
    # ==============================
    best_loss = 999999.
    es_count = 0
    pbar = tqdm(range(epochs), desc='epoch')
    for epoch in pbar:
        for train_batch in train_batches:
            acc, bce = train_step(model, train_batch, optimizer)
            train_acc.update_state(acc)
            train_bce.update_state(bce)

        for valid_batch in valid_batches:
            acc, bce = valid_step(model, valid_batch)
            valid_acc.update_state(acc)
            valid_bce.update_state(bce)

        save_scalar(train_writter, train_acc, epoch, name='accuracy')
        save_scalar(valid_writter, valid_acc, epoch, name='accuracy')
        save_scalar(train_writter, train_bce, epoch, name='xentropy')
        save_scalar(valid_writter, valid_bce, epoch, name='xentropy')

        if valid_bce.result() < best_loss:
            best_loss = valid_bce.result()
            es_count = 0.
            model.save_weights(os.path.join(exp_path, 'weights'))
        else:
            es_count+=1.
        if es_count == patience:
            print('[INFO] Early Stopping Triggered')
            break


        msg = 'EPOCH {} - ES COUNT: {}/{} Train acc: {:.4f} - Val acc: {:.4f} - Train CE: {:.2f} - Val CE: {:.2f}'.format(
                                                                                      epoch,
                                                                                      es_count,
                                                                                      patience,
                                                                                      train_acc.result(),
                                                                                      valid_acc.result(),
                                                                                      train_bce.result(),
                                                                                      valid_bce.result())

        pbar.set_description(msg)

        valid_bce.reset_states()
        train_bce.reset_states()
        train_acc.reset_states()
        valid_acc.reset_states()

def get_conf(path):
    conf_file = os.path.join(path, 'conf.json')
    with open(conf_file, 'r') as handle:
        conf = json.load(handle)
        return conf

def load_weights(model, weigths):
    weights_path = '{}/weights'.format(weigths)
    model.load_weights(weights_path)
    return model

def predict(model, test_batches):
    predictions = []
    true_labels = []
    for batch in tqdm(test_batches, desc='test'):
        acc, ce, y_pred, y_true = valid_step(model, batch, return_pred=True)
        if len(y_pred.shape)>2:
            predictions.append(y_pred[:, -1, :])
        else:
            predictions.append(y_pred)

        true_labels.append(y_true)

    y_pred = tf.concat(predictions, 0)
    y_true = tf.concat(true_labels, 0)
    pred_labels = tf.argmax(y_pred, 1)

    precision, \
    recall, \
    f1, _ = precision_recall_fscore_support(y_true,
                                            pred_labels,
                                            average='macro')
    acc = accuracy_score(y_true, pred_labels)
    results = {'f1': f1,
               'recall': recall,
               'precision': precision,
               'accuracy':acc,
               'y_true':y_true,
               'y_pred':pred_labels}

    return results


def predict_from_path(path, test_batches, mode=0, save=False):
    conf_rnn = get_conf(path)

    if mode == 0:
        clf = get_lstm_attention(conf_rnn['units'],
                               conf_rnn['num_classes'],
                               conf_rnn['w'],
                                     conf_rnn['dropout'])
    if mode == 1:
        clf = get_fc_attention(conf_rnn['units'],
                                     conf_rnn['num_classes'],
                                     conf_rnn['w'])
    if mode == 2:
        clf = get_lstm_no_attention(conf_rnn['units'],
                                    conf_rnn['num_classes'],
                                    conf_rnn['max_obs'],
                                    conf_rnn['dropout'])

    clf = load_weights(clf, path)
    results = predict(clf, test_batches)

    return results



import tensorflow as tf

from core.attention import MultiHeadAttention
from core.positional import positional_encoding
from core.masking import reshape_mask

def point_wise_feed_forward_network(d_model, dff):
    return tf.keras.Sequential([
        tf.keras.layers.Dense(dff, activation='tanh'),  # (batch_size, seq_len, dff)
        tf.keras.layers.Dense(d_model)  # (batch_size, seq_len, d_model)
    ])

class EncoderLayer(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads, dff, rate=0.1, use_leak=False, **kwargs):
        super(EncoderLayer, self).__init__(**kwargs)

        self.mha = MultiHeadAttention(d_model, num_heads)
        self.ffn = point_wise_feed_forward_network(d_model, dff)

        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-6)

        self.use_leak = use_leak
        if use_leak:
            self.reshape_leak_1 = tf.keras.layers.Dense(d_model)
            self.reshape_leak_2 = tf.keras.layers.Dense(d_model)

        self.dropout1 = tf.keras.layers.Dropout(rate)
        self.dropout2 = tf.keras.layers.Dropout(rate)

    def call(self, x, training, mask):
        attn_output, _ = self.mha(x, mask)  # (batch_size, input_seq_len, d_model)
        attn_output = self.dropout1(attn_output, training=training)

        if self.use_leak:
            out1 = self.layernorm1(self.reshape_leak_1(x) + attn_output)  # (batch_size, input_seq_len, d_model)
        else:
            out1 = self.layernorm1(attn_output)

        ffn_output = self.ffn(out1)  # (batch_size, input_seq_len, d_model)
        ffn_output = self.dropout2(ffn_output, training=training)

        if self.use_leak:
            out2 = self.layernorm2(self.reshape_leak_2(out1) + ffn_output) # (batch_size, input_seq_len, d_model)
        else:
            out2 = self.layernorm2(ffn_output)

        return out2


class Encoder(tf.keras.layers.Layer):
    def __init__(self, num_layers, d_model, num_heads, dff,
                 base=10000, rate=0.1, use_leak=False, **kwargs):
        super(Encoder, self).__init__(**kwargs)

        self.d_model = d_model
        self.num_layers = num_layers
        self.base = base
        self.inp_transform = tf.keras.layers.Dense(d_model)
        self.enc_layers = [EncoderLayer(d_model, num_heads, dff, rate, use_leak)
                            for _ in range(num_layers)]
        self.dropout = tf.keras.layers.Dropout(rate)

    def call(self, data, training=False):
        # adding embedding and position encoding.
        x_pe = positional_encoding(data['times'], self.d_model, mjd=True)
        # x_pe = self.pe_emb(data['times'])

        x_transformed = self.inp_transform(data['input'])

        transformed_input = x_transformed + x_pe
        x = self.dropout(transformed_input, training=training)

        for i in range(self.num_layers):
            x = self.enc_layers[i](x, training, data['mask_in'])

        return x  # (batch_size, input_seq_len, d_model)



import multiprocessing as mp
import tensorflow as tf
import pandas as pd
import numpy as np
import logging
import os

from core.masking import get_masked, set_random, get_padding_mask
from core.utils import standardize
from joblib import Parallel, delayed
from tqdm import tqdm

from time import time
from joblib import wrap_non_picklable_objects

logging.getLogger('tensorflow').setLevel(logging.ERROR)  # suppress warnings

def _bytes_feature(value):
    """Returns a bytes_list from a string / byte."""
    if isinstance(value, type(tf.constant(0))):
        value = value.numpy() # BytesList won't unpack a string from an EagerTensor.
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

def _float_feature(list_of_floats):  # float32
    return tf.train.Feature(float_list=tf.train.FloatList(value=list_of_floats))

def _int64_feature(value):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))

def get_example(lcid, label, lightcurve):
    """
    Create a record example from numpy values.

    Args:
        lcid (string): object id
        label (int): class code
        lightcurve (numpy array): time, magnitudes and observational error

    Returns:
        tensorflow record
    """

    f = dict()

    dict_features={
    'id': tf.train.Feature(bytes_list=tf.train.BytesList(value=[str(lcid).encode()])),
    'label': tf.train.Feature(int64_list=tf.train.Int64List(value=[label])),
    'length': tf.train.Feature(int64_list=tf.train.Int64List(value=[lightcurve.shape[0]])),
    }
    element_context = tf.train.Features(feature = dict_features)

    dict_sequence = {}
    for col in range(lightcurve.shape[1]):
        seqfeat = _float_feature(lightcurve[:, col])
        seqfeat = tf.train.FeatureList(feature = [seqfeat])
        dict_sequence['dim_{}'.format(col)] = seqfeat

    element_lists = tf.train.FeatureLists(feature_list=dict_sequence)
    ex = tf.train.SequenceExample(context = element_context,
                                  feature_lists= element_lists)
    return ex

def divide_training_subset(frame, train, val):
    """
    Divide the dataset into train, validation and test subsets.
    Notice that:
        test = 1 - (train + val)

    Args:
        frame (Dataframe): Dataframe following the astro-standard format
        dest (string): Record destination.
        train (float): train fraction
        val (float): validation fraction
    Returns:
        tuple x3 : (name of subset, subframe with metadata)
    """

    frame = frame.sample(frac=1)
    n_samples = frame.shape[0]
    n_train = int(n_samples*train)
    n_val = int(n_samples*val//2)

    sub_train = frame.iloc[:n_train]
    sub_val   = frame.iloc[n_train:n_train+n_val]
    sub_test  = frame.iloc[n_train+n_val:]

    return ('train', sub_train), ('val', sub_val), ('test', sub_test)

def process_lc(row, source, unique_classes, writer):
    """
    Filter a sample according to our astronomical criteria and write
    example in the corresponding record.

    Args:
        row (Pandas serie): Dataframe row containing information about
                            a particular sample
        source (string):  Lightcurves folder path
        unique_classes (list): list with unique classes
        writer (tf.writer) : record writer

    """

    path  = row['Path'].split('/')[-1]
    label = list(unique_classes).index(row['Class'])
    lc_path = os.path.join(source, path)
    try:
        observations = pd.read_csv(lc_path)

        observations.columns = ['mjd', 'mag', 'errmag']
        observations = observations.dropna()
        observations.sort_values('mjd')
        observations = observations.drop_duplicates(keep='last')
        numpy_lc = observations.values
        ex = get_example(row['ID'], label, numpy_lc)
        writer.write(ex.SerializeToString())
    except:
        print('[ERROR] {} Lightcurve could not be processed.'.format(row['ID']))

@wrap_non_picklable_objects
def process_lc2(row, source, unique_classes, max_obs):
    path  = row['Path'].split('/')[-1]
    label = list(unique_classes).index(row['Class'])
    lc_path = os.path.join(source, path)
    
    observations = pd.read_csv(lc_path, delim_whitespace=True, names=['mjd', 'mag', 'std'])
    
    observations.columns = ['mjd', 'mag', 'errmag']
    observations = observations.dropna()
    observations.sort_values('mjd')
    observations = observations.drop_duplicates(keep='last')

    if max_obs != -1:
        length = observations.shape[0]
        rest = length%max_obs
        filler = max_obs - rest
        observations = observations.reindex(range(length+filler), fill_value=0)
    
    
    numpy_lc = observations.values

    return row['ID'], label, numpy_lc

def process_lc3(lc_index, label, numpy_lc, writer):
    try:
        ex = get_example(lc_index, label, numpy_lc)
        writer.write(ex.SerializeToString())
    except:
        print('[INFO] {} could not be processed'.format(lc_index))

def write_records(frame, dest, max_lcs_per_record, source, unique, n_jobs=None, max_obs=200):
    # Get frames with fixed number of lightcurves
    collection = [frame.iloc[i:i+max_lcs_per_record] \
                  for i in range(0, frame.shape[0], max_lcs_per_record)]

#     if 'test' in dest:
#         max_obs = 200 # For testing we split the entire lc in windows of 200 obs
        
    # Iterate over subset
    # First process and then serialize
    for counter, subframe in enumerate(collection):
        var = Parallel(n_jobs=n_jobs)(delayed(process_lc2)(row, source, unique, max_obs) \
                                    for k, row in subframe.iterrows())

        with tf.io.TFRecordWriter(dest+'/chunk_{}.record'.format(counter)) as writer:
            for counter2, data_lc in enumerate(var):
#                 if 'test' in dest:
                oid, y, x = data_lc
                windows = np.split(x, x.shape[0]/max_obs)
                for w in windows:
                    if np.sum(w) != 0:
                        process_lc3(oid, y, w, writer)
#                 else:
#                     process_lc3(*data_lc, writer)



def create_dataset(meta_df,
                   source='data/raw_data/macho/MACHO/LCs',
                   target='data/records/macho/',
                   n_jobs=None,
                   subsets_frac=(0.5, 0.25),
                   max_lcs_per_record=100):
    os.makedirs(target, exist_ok=True)

    bands = meta_df['Band'].unique()
    if len(bands) > 1:
        b = input('Filters {} were found. Type one to continue'.format(' and'.join(bands)))
        meta_df = meta_df[meta_df['Band'] == b]

    unique, counts = np.unique(meta_df['Class'], return_counts=True)
    info_df = pd.DataFrame()
    info_df['label'] = unique
    info_df['size'] = counts
    info_df.to_csv(os.path.join(target, 'objects.csv'), index=False)

    # Separate by class
    cls_groups = meta_df.groupby('Class')

    for cls_name, cls_meta in tqdm(cls_groups, total=len(cls_groups)):
        subsets = divide_training_subset(cls_meta,
                                         train=subsets_frac[0],
                                         val=subsets_frac[0])

        for subset_name, frame in subsets:
            dest = os.path.join(target, subset_name, cls_name)
            os.makedirs(dest, exist_ok=True)
            write_records(frame, dest, max_lcs_per_record, source, unique, n_jobs)

def get_sample(sample):
    """
    Read a serialized sample and convert it to tensor
    Context and sequence features should match with the name used when writing.
    Args:
        sample (binary): serialized sample

    Returns:
        type: decoded sample
    """
    context_features = {'label': tf.io.FixedLenFeature([],dtype=tf.int64),
                        'length': tf.io.FixedLenFeature([],dtype=tf.int64),
                        'id': tf.io.FixedLenFeature([], dtype=tf.string)}
    sequence_features = dict()
    for i in range(3):
        sequence_features['dim_{}'.format(i)] = tf.io.VarLenFeature(dtype=tf.float32)

    context, sequence = tf.io.parse_single_sequence_example(
                            serialized=sample,
                            context_features=context_features,
                            sequence_features=sequence_features
                            )

    input_dict = dict()
    input_dict['lcid']   = tf.cast(context['id'], tf.string)
    input_dict['length'] = tf.cast(context['length'], tf.int32)
    input_dict['label']  = tf.cast(context['label'], tf.int32)

    casted_inp_parameters = []
    for i in range(3):
        seq_dim = sequence['dim_{}'.format(i)]
        seq_dim = tf.sparse.to_dense(seq_dim)
        seq_dim = tf.cast(seq_dim, tf.float32)
        casted_inp_parameters.append(seq_dim)


    sequence = tf.stack(casted_inp_parameters, axis=2)[0]
    input_dict['input'] = sequence
    return input_dict

def sample_lc(sequence, max_obs):
    '''
    Sample a random window of "max_obs" observations from the input sequence
    '''
    serie_len = tf.shape(sequence)[0]
    curr_max_obs = tf.minimum(serie_len, max_obs)
    
    pivot = 0
    if tf.greater(serie_len, max_obs):
        pivot = tf.random.uniform([],
                                  minval=0,
                                  maxval=serie_len-curr_max_obs,
                                  dtype=tf.int32)

        sequence = tf.slice(sequence, [pivot,0], [curr_max_obs, -1])
    else:
        sequence = tf.slice(sequence, [0,0], [curr_max_obs, -1])

    return sequence

def _parse_pt(sample, msk_prob, rnd_prob, same_prob, max_obs):
    '''
    Pretraining formater
    '''

    input_dict = get_sample(sample)
    input_dict_v2 = {}
    sequence = sample_lc(input_dict['input'], max_obs)
    
    pad_mask = tf.reduce_sum(sequence, 1)
    pad_mask = tf.math.divide_no_nan(pad_mask, pad_mask)
    curr_max_obs = tf.reduce_sum(pad_mask, 0)
    pad_mask = tf.expand_dims(pad_mask, 1)
    
    sequence, mean = standardize(sequence, return_mean=True)

    seq_time = tf.slice(sequence, [0, 0], [max_obs, 1])
    seq_magn = tf.slice(sequence, [0, 1], [max_obs, 1])
    seq_errs = tf.slice(sequence, [0, 2], [max_obs, 1])


    # Save the true values
    orig_magn = seq_magn

    # [MASK] values
    mask_out = get_masked(seq_magn, msk_prob)

    # [MASK] -> Same values
    seq_magn, mask_in = set_random(seq_magn,
                                   mask_out,
                                   seq_magn,
                                   same_prob,
                                   name='set_same')

    # [MASK] -> Random value
    seq_magn, mask_in = set_random(seq_magn,
                                   mask_in,
                                   tf.random.shuffle(seq_magn),
                                   rnd_prob,
                                   name='set_random')

    time_steps = tf.shape(seq_magn)[0]

    mask_out = tf.reshape(mask_out, [max_obs, 1])
    mask_out = tf.multiply(mask_out, pad_mask)
    
    mask_in = tf.reshape(mask_in, [max_obs, 1])
    mask_in = tf.maximum(1.-pad_mask, mask_in)

#     if curr_max_obs < max_obs:
#         print('Masking')
#         filler    = tf.ones([max_obs-curr_max_obs, 1])
#         mask_in   = tf.concat([mask_in, filler], 0)
#         seq_magn  = tf.concat([seq_magn, 1.-filler], 0)
#         seq_time  = tf.concat([seq_time, 1.-filler], 0)
#         mask_out  = tf.concat([mask_out, 1.-filler], 0)
#         orig_magn = tf.concat([orig_magn, 1.-filler], 0)

    input_dict['output']   = orig_magn
    input_dict['input']    = seq_magn
    input_dict['times']    = seq_time
    input_dict['mask_out'] = mask_out
    input_dict['mask_in']  = mask_in
    input_dict['length']   = curr_max_obs
    input_dict['mean']     = mean
    input_dict['obserr']   = seq_errs

    return input_dict

def _parse_normal(sample, max_obs):
    '''
    Specific task formater
    '''
    input_dict = get_sample(sample)
       
    sequence = sample_lc(input_dict['input'], max_obs)
    
    mask = tf.reduce_sum(sequence, 1)
    curr_max_obs = tf.reduce_sum(mask, 0)
    mask = 1.-tf.math.divide_no_nan(mask, mask)
    
    sequence = standardize(sequence)
    seq_time = tf.slice(sequence, [0, 0], [max_obs, 1])
    seq_magn = tf.slice(sequence, [0, 1], [max_obs, 1])
    seq_errs = tf.slice(sequence, [0, 2], [max_obs, 1])

    mask = tf.reshape(mask, [max_obs, 1])

    input_dict['input']    = seq_magn
    input_dict['times']    = seq_time
    input_dict['obserr']   = seq_errs
    input_dict['mask_in']  = mask
    input_dict['length']   = curr_max_obs

    return input_dict

def adjust_fn(func, msk_prob, rnd_prob, same_prob, max_obs):
    def wrap(*args, **kwargs):
        result = func(*args, msk_prob, rnd_prob, same_prob, max_obs)
        return result
    return wrap

def pretraining_records(source, batch_size, shuffle=False, max_obs=100,
                        msk_frac=0.2, rnd_frac=0.1, same_frac=0.1):
    """
    Pretraining data loader.
    This method build the ASTROMER input format.
    ASTROMER format is based on the BERT masking strategy.

    Args:
        source (string): Record folder
        batch_size (int): Batch size
        no_shuffle (bool): Do not shuffle training and validation dataset
        max_obs (int): Max. number of observation per serie
        msk_frac (float): fraction of values to be predicted ([MASK])
        rnd_frac (float): fraction of [MASKED] values to replace with random values
        same_frac (float): fraction of [MASKED] values to replace with true values

    Returns:
        Tensorflow Dataset: Iterator withg preprocessed batches
    """
    fn = adjust_fn(_parse_pt, msk_frac, rnd_frac, same_frac, max_obs)

    rec_paths = [os.path.join(source, folder, x) for folder in os.listdir(source) \
                 for x in os.listdir(os.path.join(source, folder))]

    dataset = tf.data.TFRecordDataset(rec_paths)
    if shuffle:
        print('[INFO] Shuffling')
        dataset = dataset.shuffle(10000)

    dataset = dataset.map(fn)
    dataset = dataset.cache()
    dataset = dataset.padded_batch(batch_size)
    dataset = dataset.prefetch(buffer_size=tf.data.AUTOTUNE)
    return dataset

def adjust_fn_clf(func, max_obs):
    def wrap(*args, **kwargs):
        result = func(*args, max_obs)
        return result
    return wrap

def clf_records(source, batch_size, max_obs=100, take=1):
    """
    Classification data loader.
    It creates ASTROMER-like input but without masking

    Args:
        source (string): Record folder
        batch_size (int): Batch size
        max_obs (int): Max. number of observation per serie
        take (int):  Number of batches.
                     If 'take' is -1 then it returns the whole dataset without
                     shuffle and oversampling (i.e., testing case)
    Returns:
        Tensorflow Dataset: Iterator withg preprocessed batches
    """

    fn = adjust_fn_clf(_parse_normal, max_obs)
    rec_paths = [os.path.join(source, folder, x) for folder in os.listdir(source) \
                 for x in os.listdir(os.path.join(source, folder))]

    if take < 0:
        print('[INFO] No shuffle No Oversampling'.format(take))
        dataset = tf.data.TFRecordDataset(rec_paths)
        dataset = dataset.map(fn).cache()
        dataset = dataset.batch(batch_size)
        dataset = dataset.prefetch(buffer_size=tf.data.AUTOTUNE)
        return dataset
    else:
        print('[INFO] Taking {} balanced batches'.format(take))
        datasets = [tf.data.TFRecordDataset(x) for x in rec_paths]
        datasets = [dataset.repeat() for dataset in datasets]
        dataset = tf.data.experimental.sample_from_datasets(datasets)
        dataset = dataset.map(fn)
        dataset = dataset.batch(batch_size)
        dataset = dataset.take(take)
        # dataset = dataset.cache()
        dataset = dataset.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)
        return dataset

# SINUSOIDAL DATA
def count(stop):
    input_dict = dict()
    while True:
        seq_times = tf.reshape(tf.linspace(0., 20., stop), [stop, 1])
        seq_times = tf.cast(seq_times, tf.float32)
        seq_magn  = tf.math.sin(seq_times)
        seq_magn = tf.cast(seq_magn, tf.float32)
        seq_err   = tf.random.normal([stop, 1], mean=0, stddev=1)
        seq_err = tf.cast(seq_err, tf.float32)

        # Save the true values
        orig_magn = seq_magn

        # [MASK] values
        mask_out = get_masked(seq_magn, 0.5)

        # [MASK] -> Same values
        seq_magn, mask_in = set_random(seq_magn,
                                       mask_out,
                                       seq_magn,
                                       0.2,
                                       name='set_same')

        # [MASK] -> Random value
        seq_magn, mask_in = set_random(seq_magn,
                                       mask_in,
                                       tf.random.shuffle(seq_magn),
                                       0.2,
                                       name='set_random')

        time_steps = tf.shape(seq_magn)[0]

        mask_out = tf.reshape(mask_out, [time_steps, 1])
        mask_in = tf.reshape(mask_in, [time_steps, 1])

        input_dict['input']    = seq_magn
        input_dict['output']   = orig_magn
        input_dict['times']    = seq_times
        input_dict['obserr']   = tf.zeros_like(seq_magn)
        input_dict['mask_in']  = mask_in
        input_dict['mask_out'] = mask_out
        input_dict['length']   = time_steps

        yield input_dict



def from_generator(len, n, batch_size):

    ds_counter = tf.data.Dataset.from_generator(count,
                                                args=[len],
                                                output_types={'input': tf.float32,
                                                              'output': tf.float32,
                                                              'times': tf.float32,
                                                              'obserr': tf.float32,
                                                              'mask_in': tf.float32,
                                                              'mask_out': tf.float32,
                                                              'length': tf.int32
                                                              },
                                                output_shapes ={'input': (len, 1),
                                                                'output': (len, 1),
                                                                'times': (len, 1),
                                                                'obserr': (len, 1),
                                                                'mask_in': (len, 1),
                                                                'mask_out': (len, 1),
                                                                'length': ()
                                                                })



    ds_counter = ds_counter.take(n).cache()
    ds_counter = ds_counter.batch(batch_size)
    return ds_counter



def _parse_pt_v2(sample, msk_prob, rnd_prob, same_prob, max_obs):
    '''
    Pretraining formater
    '''

    input_dict = get_sample(sample)
    input_dict_v2 = {}
    sequence = sample_lc(input_dict['input'], max_obs)
    
    pad_mask = tf.reduce_sum(sequence, 1)
    pad_mask = tf.math.divide_no_nan(pad_mask, pad_mask)
    curr_max_obs = tf.reduce_sum(pad_mask, 0)
    pad_mask = tf.expand_dims(pad_mask, 1)
    
    sequence, mean = standardize(sequence, return_mean=True)

    seq_time = tf.slice(sequence, [0, 0], [max_obs, 1])
    seq_magn = tf.slice(sequence, [0, 1], [max_obs, 1])
    seq_errs = tf.slice(sequence, [0, 2], [max_obs, 1])


    # Save the true values
    orig_magn = seq_magn

    # [MASK] values
    mask_out = get_masked(seq_magn, msk_prob)

    # [MASK] -> Same values
    seq_magn, mask_in = set_random(seq_magn,
                                   mask_out,
                                   seq_magn,
                                   same_prob,
                                   name='set_same')

    # [MASK] -> Random value
    seq_magn, mask_in = set_random(seq_magn,
                                   mask_in,
                                   tf.random.shuffle(seq_magn),
                                   rnd_prob,
                                   name='set_random')

    time_steps = tf.shape(seq_magn)[0]

    mask_out = tf.reshape(mask_out, [max_obs, 1])
    mask_out = tf.multiply(mask_out, pad_mask)
    
    mask_in = tf.reshape(mask_in, [max_obs, 1])
    mask_in = tf.maximum(1.-pad_mask, mask_in)



    # output_dict['output']   = orig_magn
    input_dict_v2['input']    = seq_magn
    input_dict_v2['times']    = seq_time
    # input_dict['mask_out'] = mask_out
    input_dict_v2['mask_in']  = mask_in
    # input_dict['length']   = curr_max_obs
    # input_dict['mean']     = mean
    # input_dict['obserr']   = seq_errs

    return input_dict_v2, input_dict['label']


def pretraining_records_v2(source, batch_size, shuffle=False, max_obs=100,
                        msk_frac=0.2, rnd_frac=0.1, same_frac=0.1):
    """
    Pretraining data loader.
    This method build the ASTROMER input format.
    ASTROMER format is based on the BERT masking strategy.

    Args:
        source (string): Record folder
        batch_size (int): Batch size
        no_shuffle (bool): Do not shuffle training and validation dataset
        max_obs (int): Max. number of observation per serie
        msk_frac (float): fraction of values to be predicted ([MASK])
        rnd_frac (float): fraction of [MASKED] values to replace with random values
        same_frac (float): fraction of [MASKED] values to replace with true values

    Returns:
        Tensorflow Dataset: Iterator withg preprocessed batches
    """
    fn = adjust_fn(_parse_pt_v2, msk_frac, rnd_frac, same_frac, max_obs)

    rec_paths = [os.path.join(source, folder, x) for folder in os.listdir(source) \
                 for x in os.listdir(os.path.join(source, folder))]

    dataset = tf.data.TFRecordDataset(rec_paths)
    if shuffle:
        print('[INFO] Shuffling')
        dataset = dataset.shuffle(10000)

    dataset = dataset.map(fn)
    dataset = dataset.cache()
    dataset = dataset.padded_batch(batch_size)
    dataset = dataset.prefetch(buffer_size=tf.data.AUTOTUNE)
    return dataset


def _parse_pt_numpy(sample, msk_prob, rnd_prob, same_prob, max_obs):
    '''
    Pretraining formater
    '''

    input_dict = {}
    sequence = sample_lc(sample, max_obs)
    
    pad_mask = tf.reduce_sum(sequence, 1)
    pad_mask = tf.math.divide_no_nan(pad_mask, pad_mask)
    curr_max_obs = tf.reduce_sum(pad_mask, 0)
    pad_mask = tf.expand_dims(pad_mask, 1)
    pad_mask = tf.cast(pad_mask, tf.float32)

    sequence, mean = standardize(sequence, return_mean=True)

    seq_time = tf.slice(sequence, [0, 0], [max_obs, 1])
    seq_time = tf.cast(seq_time, tf.float32)
    seq_magn = tf.slice(sequence, [0, 1], [max_obs, 1])
    seq_magn = tf.cast(seq_magn, tf.float32)
    seq_errs = tf.slice(sequence, [0, 2], [max_obs, 1])
    seq_errs = tf.cast(seq_errs, tf.float32)



    # # Save the true values
    orig_magn = seq_magn

    # [MASK] values
    mask_out = get_masked(seq_magn, msk_prob)

    # [MASK] -> Same values
    seq_magn, mask_in = set_random(seq_magn,
                                   mask_out,
                                   seq_magn,
                                   same_prob,
                                   name='set_same')

    # [MASK] -> Random value
    seq_magn, mask_in = set_random(seq_magn,
                                   mask_in,
                                   tf.random.shuffle(seq_magn),
                                   rnd_prob,
                                   name='set_random')

    # time_steps = tf.shape(seq_magn)[0]

    mask_out = tf.reshape(mask_out, [max_obs, 1])
    mask_out = tf.multiply(mask_out, pad_mask)
    
    mask_in = tf.reshape(mask_in, [max_obs, 1])
    mask_in = tf.maximum(1.-pad_mask, mask_in)

    #     if curr_max_obs < max_obs:
    #         print('Masking')
    #         filler    = tf.ones([max_obs-curr_max_obs, 1])
    #         mask_in   = tf.concat([mask_in, filler], 0)
    #         seq_magn  = tf.concat([seq_magn, 1.-filler], 0)
    #         seq_time  = tf.concat([seq_time, 1.-filler], 0)
    #         mask_out  = tf.concat([mask_out, 1.-filler], 0)
    #         orig_magn = tf.concat([orig_magn, 1.-filler], 0)

    input_dict['output']   = orig_magn
    input_dict['input']    = seq_magn
    input_dict['times']    = seq_time
    input_dict['mask_out'] = mask_out
    input_dict['mask_in']  = mask_in
    input_dict['length']   = curr_max_obs
    # input_dict['mean']     = mean
    # input_dict['obserr']   = seq_errs

    return input_dict
def pretraining_numpy(arr, batch_size, shuffle=False, max_obs=100,
                        msk_frac=0.2, rnd_frac=0.1, same_frac=0.1):
    """
    Pretraining data loader.
    This method build the ASTROMER input format.
    ASTROMER format is based on the BERT masking strategy.

    Args:
        source (string): Record folder
        batch_size (int): Batch size
        no_shuffle (bool): Do not shuffle training and validation dataset
        max_obs (int): Max. number of observation per serie
        msk_frac (float): fraction of values to be predicted ([MASK])
        rnd_frac (float): fraction of [MASKED] values to replace with random values
        same_frac (float): fraction of [MASKED] values to replace with true values

    Returns:
        Tensorflow Dataset: Iterator withg preprocessed batches
    """

    
    fn = adjust_fn(_parse_pt_numpy, msk_frac, rnd_frac, same_frac, max_obs)



    dataset = tf.data.Dataset.from_tensor_slices(arr)
    if shuffle:
        print('[INFO] Shuffling')
        dataset = dataset.shuffle(1000)

    dataset = dataset.map(fn)
    dataset = dataset.cache()
    dataset = dataset.padded_batch(batch_size)
    dataset = dataset.prefetch(buffer_size=tf.data.AUTOTUNE)
    return dataset


import tensorflow as tf
import logging
import os, sys

from core.output    import RegLayer
from core.tboard    import save_scalar, draw_graph
from core.losses    import custom_rmse, custom_bce
from core.metrics   import custom_acc
from core.encoder   import Encoder
from core.utils     import load_weights
from tensorflow.keras.layers import Input, Dense
from tensorflow.keras.optimizers import Adam
from tensorflow.keras import Model
from tqdm import tqdm
import time
import git
import tempfile
import shutil
import json

logging.getLogger('tensorflow').setLevel(logging.ERROR)  # suppress warnings
os.system('clear')

class ASTROMER:
    def __init__(self, num_layers=2,
                 d_model=200,
                 num_heads=2,
                 dff=256,
                 base=10000,
                 dropout=0.1,
                 use_leak=False,
                 no_train=True,
                 maxlen=100,
                 batch_size=None,
                 pretrained_weights=False,
                 finetuning = None,
                 overwrite=False):
        self.num_layers=num_layers
        self.d_model=d_model
        self.num_heads=num_heads
        self.dff=dff
        self.base=base
        self.dropout=dropout
        self.use_leak=use_leak
        self.no_train=no_train
        self.maxlen=maxlen
        self.batch_size=batch_size
        self.pretrained_weights=pretrained_weights
        self.overwrite_weights=overwrite
        
        self.git_link="https://github.com/HarshVardhanGoyal/test_model.git"
        self.model_path = os.path.join(os.getcwd(),"weights")
        self.model_name = "astromer_10022021"

        # Getting the desired model_name
        if finetuning in [None, "ogle", "alcock", "ogle_100"]:
            self.finetuning = finetuning
        else:
            raise ValueError("Specified fine-tuned model does not exist")
        
        if self.pretrained_weights:
            self.load_configuration()

        serie  = Input(shape=(self.maxlen, 1),
                    batch_size=None,
                    name='input')
        times  = Input(shape=(self.maxlen, 1),
                    batch_size=None,
                    name='times')
        mask   = Input(shape=(self.maxlen, 1),
                    batch_size=None,
                    name='mask')
        length = Input(shape=(self.maxlen,),
                    batch_size=None,
                    dtype=tf.int32,
                    name='length')

        placeholder = {'input':serie,
                    'mask_in':mask,
                    'times':times,
                    'length':length}

        encoder = Encoder(self.num_layers,
                    self.d_model,
                    self.num_heads,
                    self.dff,
                    base=self.base,
                    rate=self.dropout,
                    use_leak=self.use_leak,
                    name='encoder')

        if self.no_train:
            encoder.trainable = False

        x = encoder(placeholder)

        x = RegLayer(name='regression')(x)
        self.model = Model(inputs=placeholder,
                    outputs=x,
                    name="ASTROMER")

        if self.pretrained_weights:
            # Loading the model weights
            try:
                if self.finetuning is not None:
                    print(f"Loading {self.finetuning} finetuned Model...")
                    self.model.load_weights(os.path.join(self.model_path, self.model_name, "finetuning", self.finetuning, "weights"))
                else:
                    print("Loading Base Model...")
                    self.model.load_weights(os.path.join(self.model_path, self.model_name, "weights"))
                print("Weights loaded successfully!")
            except:
                print("Couldn't load model weights!") 
    
    # Function to get the model
    def get_model(self):

        return self.model

    def load_configuration(self):

        # model_name="astromer_10022021"
        # git_link="https://github.com/HarshVardhanGoyal/test_model.git"
        # model_path = os.path.join(os.getcwd(),"weights")

        # Creating a temporary directory and cloning the desired repo
        test_dir = tempfile.mkdtemp()
        git.Repo.clone_from(self.git_link, test_dir, branch='main', depth=1)
        # time.sleep(2)

        # Defining the source path of the weights (i.e temporary directory)
        sourcepath = os.path.join(test_dir, self.model_name)

        # Checking the directories
        if not os.path.exists(self.model_path):
            os.makedirs(self.model_path)
        
        if os.path.exists(os.path.join(self.model_path, self.model_name)):
            if self.overwrite_weights:
                shutil.rmtree(os.path.join(self.model_path, self.model_name))
                shutil.move(sourcepath,self.model_path)
            else:
                print("The saved weights already exists")
        else:
            # Moving the weights to the code directory
            shutil.move(sourcepath,self.model_path)
     
        # Removing the temporary directory
        os.system('rmdir /S /Q "{}"'.format(test_dir))

        # Loading the conf file from the downloaded weights folder
        conf_file = os.path.join( self.model_path, self.model_name, 'conf.json')
        with open(conf_file, 'r') as handle:
            conf = json.load(handle)

        # Getting the astromer model as per the desired configuration
        self.num_layers=conf['layers']
        self.d_model   =conf['head_dim']
        self.num_heads =conf['heads']
        self.dff       =conf['dff']
        self.base      =conf['base']
        self.dropout   =conf['dropout']
        self.maxlen    =conf['max_obs']
        self.use_leak  =conf['use_leak']
        


    @tf.function
    def train_step(self, batch, opt):
        with tf.GradientTape() as tape:
            x_pred = self.model(batch)
            
            mse = custom_rmse(y_true=batch['output'],
                            y_pred=x_pred,
                            mask=batch['mask_out'])

        
        grads = tape.gradient(mse, self.model.trainable_weights)
        opt.apply_gradients(zip(grads, self.model.trainable_weights))
        return mse

    @tf.function
    def valid_step(self, batch, return_pred=False, normed=False):
        with tf.GradientTape() as tape:
            x_pred = self.model(batch)
            # if normed:
            #     mean_x = tf.reshape(batch['mean'][:, 1], [-1, 1, 1])
            #     x_true = batch['output'] + mean_x
            #     x_pred = x_pred + mean_x
            #
            #     mse = custom_rmse(y_true=x_true,
            #                       y_pred=x_pred,
            #                       mask=batch['mask_out'])
            # else:
            x_true = batch['output']
            mse = custom_rmse(y_true=x_true,
                            y_pred=x_pred,
                            mask=batch['mask_out'])

        if return_pred:
            return mse, x_pred, x_true
        return mse

    def train(self,
            train_dataset,
            valid_dataset,
            patience=20,
            exp_path='./experiments/test',
            epochs=1,
            finetuning=False,
            use_random=True,
            num_cls=2,
            lr=1e-3,
            verbose=1):

        os.makedirs(exp_path, exist_ok=True)

        # Tensorboard
        train_writter = tf.summary.create_file_writer(
                                        os.path.join(exp_path, 'logs', 'train'))
        valid_writter = tf.summary.create_file_writer(
                                        os.path.join(exp_path, 'logs', 'valid'))

        batch = [t for t in train_dataset.take(1)][0]
        draw_graph(self.model, batch, train_writter, exp_path)

        # Optimizer
        optimizer = tf.keras.optimizers.Adam(lr,
                                            beta_1=0.9,
                                            beta_2=0.98,
                                            epsilon=1e-9)
        # To save metrics
        train_mse  = tf.keras.metrics.Mean(name='train_mse')
        valid_mse  = tf.keras.metrics.Mean(name='valid_mse')

        # Training Loop
        best_loss = 999999.
        es_count = 0
        pbar = tqdm(range(epochs), desc='epoch')
        for epoch in pbar:
            for train_batch in train_dataset:
                mse = self.train_step(train_batch, optimizer)
                train_mse.update_state(mse)

            for valid_batch in valid_dataset:
                mse = self.valid_step(valid_batch)
                valid_mse.update_state(mse)

            msg = 'EPOCH {} - ES COUNT: {}/{} train mse: {:.4f} - val mse: {:.4f}'.format(epoch,
                                                                                        es_count,
                                                                                        patience,
                                                                                        train_mse.result(),
                                                                                        valid_mse.result())

            pbar.set_description(msg)

            save_scalar(train_writter, train_mse, epoch, name='mse')
            save_scalar(valid_writter, valid_mse, epoch, name='mse')


            if valid_mse.result() < best_loss:
                best_loss = valid_mse.result()
                es_count = 0.
                self.model.save_weights(exp_path+'/weights')
            else:
                es_count+=1.
            if es_count == patience:
                print('[INFO] Early Stopping Triggered')
                break

            train_mse.reset_states()
            valid_mse.reset_states()

    def predict(self,
                dataset,
                conf,
                predic_proba=False):

        total_mse, inputs, reconstructions = [], [], []
        masks, times = [], []
        for step, batch in tqdm(enumerate(dataset), desc='prediction'):
            mse, x_pred, x_true = self.valid_step(
                                            batch,
                                            return_pred=True,
                                            normed=True)

            total_mse.append(mse)
            times.append(batch['times'])
            inputs.append(x_true)
            reconstructions.append(x_pred)
            masks.append(batch['mask_out'])

        res = {'mse':tf.reduce_mean(total_mse).numpy(),
            'x_pred': tf.concat(reconstructions, 0),
            'x_true': tf.concat(inputs, 0),
            'mask': tf.concat(masks, 0),
            'time': tf.concat(times, 0)}

        return res





from setuptools import setup, find_packages


setup(
    name='test_model_hv',
    version='0.0.22',
    license='MIT',
    author="Vishnu M Harshvardhan",
    author_email='m.vishnu701@gmail.com, reachhvg@gmail.com',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    keywords='',
    py_modules=['core/*'],
    install_requires=[
        'scikit-learn',
        'pandas',
        'joblib',
        'tqdm',
        'graphviz',
        'pydot',
        'scikit-learn',
        'seaborn',
        'googledrivedownloader',
        'dask[dataframe]',
        'optuna',
        'GitPython'
      ],

)